# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Aggregate top level objects from all Scapy modules.
"""

from scapy.base_classes import *
from scapy.config import *
from scapy.dadict import *
from scapy.data import *
from scapy.error import *
from scapy.themes import *
from scapy.arch import *
from scapy.interfaces import *

from scapy.plist import *
from scapy.fields import *
from scapy.packet import *
from scapy.asn1fields import *
from scapy.asn1packet import *

from scapy.utils import *
from scapy.route import *
from scapy.sendrecv import *
from scapy.sessions import *
from scapy.supersocket import *
from scapy.volatile import *
from scapy.as_resolvers import *

from scapy.automaton import *
from scapy.autorun import *

from scapy.main import *
from scapy.consts import *
from scapy.compat import raw  # noqa: F401

from scapy.layers.all import *

from scapy.asn1.asn1 import *
from scapy.asn1.ber import *
from scapy.asn1.mib import *

from scapy.pipetool import *
from scapy.scapypipes import *

if conf.ipv6_enabled:  # noqa: F405
    from scapy.utils6 import *  # noqa: F401
    from scapy.route6 import *  # noqa: F401

from scapy.ansmachine import *



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
.NET RemoTing Protocol

This implements:
- [MS-NRTP] - .NET Remoting Core Protocol
- [MS-NRBF] - .NET Remoting Binary Format
"""

import enum
import functools
import struct

from scapy.automaton import Automaton, ATMT
from scapy.config import conf
from scapy.main import interact
from scapy.fields import (
    ByteEnumField,
    ByteField,
    ConditionalField,
    FieldLenField,
    FieldListField,
    FlagsField,
    LEIntField,
    LELongField,
    LEShortEnumField,
    LEShortField,
    LESignedIntField,
    LESignedLongField,
    LESignedShortField,
    LenField,
    MSBExtendedField,
    MultipleTypeField,
    PacketField,
    PacketListField,
    SignedByteField,
    StrField,
    StrFixedLenField,
    StrLenField,
    StrLenFieldUtf16,
)
from scapy.packet import Packet
from scapy.supersocket import StreamSocket


# [MS-NRTP] sect 2.2.3.2.1


class CountedString(Packet):
    fields_desc = [
        ByteEnumField(
            "StringEncoding",
            0,
            {
                0: "Unicode",
                1: "UTF8",
            },
        ),
        FieldLenField("Length", None, fmt="<I", count_of="StringData"),
        MultipleTypeField(
            [
                (
                    StrLenFieldUtf16(
                        "StringData", "", length_from=lambda pkt: pkt.Length
                    ),
                    lambda pkt: pkt.StringEncoding == 0,
                ),
                (
                    StrLenField("StringData", "", length_from=lambda pkt: pkt.Length),
                    lambda pkt: pkt.StringEncoding == 1,
                ),
            ],
            StrLenField("StringData", "", length_from=lambda pkt: pkt.Length),
        ),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


CountedStringField = lambda name: PacketField(name, CountedString(), CountedString)

# [MS-NRTP] sect 2.2.3.1.4

NRTP_HeaderDataFormat = {
    0: "Void",
    1: "CountedString",
    2: "Byte",
    3: "Uint16",
    4: "Int32",
}

# [MS-NRTP] sect 2.2.3.3.3.1


class NRTPHeader(Packet):
    fields_desc = [
        LEShortEnumField(
            "HeaderToken",
            0,
            {
                0: "End",
                1: "Custom",
                2: "StatusCode",
                3: "StatusPhrase",
                4: "RequestUri",
                5: "CloseConnection",
                6: "ContentType",
            },
        ),
    ]

    registered_headers = {}

    @classmethod
    def register_variant(cls, id=None):
        cls.registered_headers[cls.HeaderToken.default] = cls

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 2:
            return cls.registered_headers.get(
                struct.unpack("<H", _pkt[:2])[0], NRTPUnknownHeader
            )
        return cls

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRTP] sect 2.2.3.3.3.1


class NRTPEndHeader(NRTPHeader):
    HeaderToken = 0
    fields_desc = [
        NRTPHeader,
    ]


# [MS-NRTP] sect 2.2.3.3.3.2


class NRTPCustomHeader(NRTPHeader):
    HeaderToken = 1
    fields_desc = [
        NRTPHeader,
        CountedStringField("HeaderName"),
        CountedStringField("HeaderValue"),
    ]


# [MS-NRTP] sect 2.2.3.3.3.3


class NRTPStatusCodeHeader(NRTPHeader):
    HeaderToken = 2
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 3, NRTP_HeaderDataFormat),
        LEShortField("StatusCodeValue", 0),
    ]


# [MS-NRTP] sect 2.2.3.3.3.4


class NRTPStatusPhraseHeader(NRTPHeader):
    HeaderToken = 3
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 1, NRTP_HeaderDataFormat),
        CountedStringField("StatusPhraseValue"),
    ]


# [MS-NRTP] sect 2.2.3.3.3.5


class NRTPRequestUriHeader(NRTPHeader):
    HeaderToken = 4
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 1, NRTP_HeaderDataFormat),
        CountedStringField("UriValue"),
    ]


# [MS-NRTP] sect 2.2.3.3.3.6


class NRTPCloseConnectionHeader(NRTPHeader):
    HeaderToken = 5
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 0, NRTP_HeaderDataFormat),
    ]


# [MS-NRTP] sect 2.2.3.3.3.7


class NRTPContentTypeHeader(NRTPHeader):
    HeaderToken = 6
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 1, NRTP_HeaderDataFormat),
        CountedStringField("ContentTypeValue"),
    ]


# [MS-NRTP] sect 2.2.3.3.3.8


class NRTPUnknownHeader(NRTPHeader):
    HeaderToken = 7
    fields_desc = [
        NRTPHeader,
        ByteEnumField("DataType", 0, NRTP_HeaderDataFormat),
        MultipleTypeField(
            [
                (
                    StrFixedLenField("DataValue", b"", length=0),
                    lambda pkt: pkt.DataType == 0,
                ),
                (
                    CountedStringField("DataValue"),
                    lambda pkt: pkt.DataType == 1,
                ),
                (
                    StrFixedLenField("DataValue", b"", length=1),
                    lambda pkt: pkt.DataType == 2,
                ),
                (
                    LEShortField("DataValue", 0),
                    lambda pkt: pkt.DataType == 3,
                ),
                (
                    LEIntField("DataValue", 0),
                    lambda pkt: pkt.DataType == 4,
                ),
            ],
            StrField("DataValue", b""),
        ),
    ]


# [MS-NRTP] sect 2.2.3.3.1


class NRTPSingleMessageContent(Packet):
    name = "NRTP Single Message Content"
    fields_desc = [
        StrFixedLenField("ProtocolId", b"\x2e\x4E\x45\x54", 4),
        ByteField("MajorVersion", 1),
        ByteField("MinorVersion", 0),
        LEShortEnumField(
            "OperationType",
            0,
            {
                0: "Request",
                1: "OneWayRequest",
                2: "Reply",
            },
        ),
        LEShortEnumField(
            "ContentDistribution",
            0,
            {
                0: "Not chunked",
                1: "Chunked",
            },
        ),
        ConditionalField(
            LenField("Length", None, fmt="<I"), lambda pkt: pkt.ContentDistribution == 0
        ),
        PacketListField(
            "Headers",
            [NRTPEndHeader()],
            None,
            next_cls_cb=lambda pkt, lst, cur, remain: (
                None if isinstance(cur, NRTPEndHeader) else NRTPHeader
            ),
        ),
    ]

    @classmethod
    def tcp_reassemble(cls, data, metadata, session):
        if not data:
            return None
        # Recover Length if available
        length = metadata.get("length", None)
        if length is None and len(data) >= 14:
            cd = struct.unpack("<H", data[8:10])[0]
            if cd == 0:
                # Not chunked
                length = struct.unpack("<I", data[10:14])[0]
                metadata["length"] = length
        if length is None:
            return cls(data)
        # Parse only the header packet
        pkt = cls(data, stop_dissection_after=Packet)
        # Is it complete?
        if pkt.payload and len(pkt.payload) >= length:
            # Get content-type
            try:
                content_type = next(
                    x.ContentTypeValue.StringData
                    for x in pkt.Headers
                    if x.HeaderToken == 6
                )
                session["content_type"] = content_type
            except StopIteration:
                # Not in this packet. Do we know it from the session?
                content_type = session.get("content_type", None)
                if not content_type:
                    return pkt
            # We have a content-type. Parse it.
            if content_type == b"application/octet-stream":
                # pkt.payload is NRBF.
                pkt.payload = NRBF(bytes(pkt.payload))
            return pkt
        return None


# [MS-NRBF] .NET Remoting Binary Format


class MSBExtendedFieldLen(MSBExtendedField):
    __slots__ = FieldLenField.__slots__

    def __init__(self, name, default, length_of=None):
        FieldLenField.__init__(self, name, default, length_of=length_of)
        super(MSBExtendedFieldLen, self).__init__(name, default)

    i2m = FieldLenField.i2m


# [MS-NRBF] sect 2.1.1.6


class NRBFLengthPrefixedString(Packet):
    fields_desc = [
        MSBExtendedFieldLen("Length", None, length_of="String"),
        StrLenField("String", b"", length_from=lambda pkt: pkt.Length),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRBF] sect 2.1.1.8


class NRBFClassTypeInfo(Packet):
    fields_desc = [
        PacketField("TypeName", NRBFLengthPrefixedString(), NRBFLengthPrefixedString),
        LESignedIntField("LibraryId", 0),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRBF] sect 2.1.2.3


class PrimitiveTypeEnum(enum.IntEnum):
    Boolean = 1
    Byte = 2
    Char = 2
    Decimal = 5
    Double = 6
    Int16 = 7
    Int32 = 8
    Int64 = 9
    SByte = 10
    Single = 11
    TimeSpan = 12
    DateTime = 13
    UInt16 = 14
    UInt32 = 15
    UInt64 = 16
    Null = 17
    String = 18


# [MS-NRBF] sect 2.1.2.2


class BinaryTypeEnum(enum.IntEnum):
    Primitive = 0
    String = 1
    Object = 2
    SystemClass = 3
    Class = 4
    ObjectArray = 5
    StringArray = 6
    PrimitiveArray = 7


# [MS-NRBF] sect 2.2.2.1


class NRBFValueWithCode(Packet):
    fields_desc = [
        ByteEnumField("PrimitiveType", 0, PrimitiveTypeEnum),
        MultipleTypeField(
            [
                (ByteField("Value", 0), lambda pkt: pkt.PrimitiveType in [1, 2, 3, 4]),
                (LESignedShortField("Value", 0), lambda pkt: pkt.PrimitiveType == 7),
                (LESignedIntField("Value", 0), lambda pkt: pkt.PrimitiveType == 8),
                (LESignedLongField("Value", 0), lambda pkt: pkt.PrimitiveType == 9),
                (SignedByteField("Value", 0), lambda pkt: pkt.PrimitiveType == 10),
                (LEShortField("Value", 0), lambda pkt: pkt.PrimitiveType == 14),
                (LEIntField("Value", 0), lambda pkt: pkt.PrimitiveType == 15),
                (LELongField("Value", 0), lambda pkt: pkt.PrimitiveType == 16),
                (
                    PacketField(
                        "Value", NRBFLengthPrefixedString(), NRBFLengthPrefixedString
                    ),
                    lambda pkt: pkt.PrimitiveType == 18,
                ),
            ],
            StrFixedLenField("Value", b"", length=0),
        ),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRBF] sect 2.2.2.2


class NRBFStringValueWithCode(NRBFValueWithCode):
    PrimitiveType = 18


StringValueWithCode = lambda name: PacketField(
    name, NRBFStringValueWithCode(), NRBFStringValueWithCode
)


# [MS-NRBF] sect 2.2.2.3


class NRBFArrayOfValueWithCode(Packet):
    fields_desc = [
        FieldLenField("Length", None, fmt="<I", count_of="ListOfValueWithCode"),
        PacketListField(
            "ListOfValueWithCode",
            [],
            NRBFValueWithCode,
            count_from=lambda pkt: pkt.Length,
        ),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# Generic record type

NRBF_RecordTypeEnumeration = {
    0: "SerializedStreamHeader",
    1: "ClassWithId",
    2: "SystemClassWithMembers",
    3: "ClassWithMembers",
    4: "SystemClassWithMembersAndTypes",
    5: "ClassWithMembersAndTypes",
    6: "BinaryObjectString",
    7: "BinaryArray",
    8: "MemberPrimitiveTyped",
    9: "MemberReference",
    10: "ObjectNull",
    11: "MessageEnd",
    12: "BinaryLibrary",
    13: "ObjectNullMultiple256",
    14: "ObjectNullMultiple",
    15: "ArraySinglePrimitive",
    16: "ArraySingleObject",
    17: "ArraySingleString",
    21: "MethodCall",
    22: "MethodReturn",
}


class NRBFRecord(Packet):
    fields_desc = [
        ByteEnumField("RecordTypeEnum", 255, NRBF_RecordTypeEnumeration),
    ]

    registered_records = {}

    @classmethod
    def register_variant(cls, id=None):
        cls.registered_records[cls.RecordTypeEnum.default] = cls

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt:
            return cls.registered_records.get(_pkt[0], NRBFRecord)
        return cls

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRBF] sect 2.2.3.1

_NRBF_MessageFlags = {
    0x00000001: "NoArgs",
    0x00000002: "ArgsInline",
    0x00000004: "ArgsIsArray",
    0x00000008: "ArgsInArray",
    0x00000010: "NoContext",
    0x00000020: "ContextInline",
    0x00000040: "ContextInArray",
    0x00000080: "MethodSignatureInArray",
    0x00000100: "PropertiesInArray",
    0x00000200: "NoReturnValue",
    0x00000400: "ReturnValueVoid",
    0x00000800: "ReturnValueInline",
    0x00001000: "ReturnValueInArray",
    0x00002000: "ExceptionInArray",
    0x00008000: "GenericMethod",
}


class NRBFBinaryMethodCall(NRBFRecord):
    RecordTypeEnum = 21
    fields_desc = [
        NRBFRecord,
        FlagsField("MessageEnum", 0, -32, _NRBF_MessageFlags),
        StringValueWithCode("MethodName"),
        StringValueWithCode("TypeName"),
        ConditionalField(
            StringValueWithCode("CallContext"),
            lambda pkt: pkt.MessageEnum.ContextInline,
        ),
        ConditionalField(
            PacketField("Args", NRBFArrayOfValueWithCode(), NRBFArrayOfValueWithCode),
            lambda pkt: pkt.MessageEnum.ArgsInline,
        ),
    ]


# [MS-NRBF] sect 2.2.3.3


class NRBFBinaryMethodReturn(NRBFRecord):
    RecordTypeEnum = 22
    fields_desc = [
        NRBFRecord,
        FlagsField("MessageEnum", 0, -32, _NRBF_MessageFlags),
        ConditionalField(
            PacketField("ReturnValue", NRBFValueWithCode(), NRBFValueWithCode),
            lambda pkt: pkt.MessageEnum.ReturnValueInline,
        ),
        ConditionalField(
            StringValueWithCode("CallContext"),
            lambda pkt: pkt.MessageEnum.ContextInline,
        ),
        ConditionalField(
            PacketField("Args", NRBFArrayOfValueWithCode(), NRBFArrayOfValueWithCode),
            lambda pkt: pkt.MessageEnum.ArgsInline,
        ),
    ]


# [MS-NRBF] sect 2.3 - Class Records

# A generic packet use for Member data


def _members_cb(pkt, lst, cur, remain):
    index = len(lst) + (1 if cur is not None else 0)
    if index >= pkt.MemberCount:
        return None
    if hasattr(pkt, "BinaryTypeEnums"):
        if index < len(pkt.BinaryTypeEnums):
            typeEnum = pkt.BinaryTypeEnums[index]
            if typeEnum == BinaryTypeEnum.Primitive:
                # Get AdditionalInfo to get the matching primitive type.
                primitiveType = pkt.AdditionalInfos[
                    sum(
                        1
                        for x in pkt.BinaryTypeEnums[:index]
                        if x
                        not in [
                            BinaryTypeEnum.String,
                            BinaryTypeEnum.Object,
                            BinaryTypeEnum.ObjectArray,
                            BinaryTypeEnum.StringArray,
                        ]
                    )
                ].Value
                return functools.partial(
                    NRBFMemberPrimitiveUnTyped,
                    type=PrimitiveTypeEnum(primitiveType),
                )
    return NRBFRecord


class _NRBFMembers(Packet):
    fields_desc = [
        PacketListField(
            "Members",
            [],
            None,
            next_cls_cb=_members_cb,
        )
    ]


# [MS-NRBF] sect 2.3.1.1


class NRBFClassInfo(Packet):
    fields_desc = [
        LESignedIntField("ObjectId", 0),
        PacketField("Name", NRBFLengthPrefixedString(), NRBFLengthPrefixedString),
        FieldLenField("MemberCount", None, fmt="<i", count_of="MemberNames"),
        PacketListField(
            "MemberNames",
            [],
            NRBFLengthPrefixedString,
            count_from=lambda pkt: pkt.MemberCount,
        ),
    ]


# [MS-NRBF] sect 2.3.1.2


class NRBFAdditionalInfo(Packet):
    __slots__ = ["type"]

    fields_desc = [
        MultipleTypeField(
            [
                (
                    ByteEnumField("Value", 0, PrimitiveTypeEnum),
                    lambda pkt: pkt.type
                    in [
                        BinaryTypeEnum.Primitive,
                        BinaryTypeEnum.PrimitiveArray,
                    ],
                ),
                (
                    PacketField(
                        "Value", NRBFLengthPrefixedString(), NRBFLengthPrefixedString
                    ),
                    lambda pkt: pkt.type == BinaryTypeEnum.SystemClass,
                ),
                (
                    PacketField("Value", NRBFClassTypeInfo(), NRBFClassTypeInfo),
                    lambda pkt: pkt.type == BinaryTypeEnum.Class,
                ),
            ],
            StrFixedLenField("Value", b"", length=0),
        ),
    ]

    def __init__(self, _pkt=None, **kwargs):
        self.type = kwargs.pop("type", BinaryTypeEnum.Primitive)
        assert isinstance(self.type, BinaryTypeEnum)
        super(NRBFAdditionalInfo, self).__init__(_pkt, **kwargs)

    def clone_with(self, *args, **kwargs):
        pkt = super(NRBFAdditionalInfo, self).clone_with(*args, **kwargs)
        pkt.type = self.type
        return pkt

    def copy(self):
        pkt = super(NRBFAdditionalInfo, self).copy()
        pkt.type = self.type
        return pkt

    def default_payload_class(self, payload):
        return conf.padding_layer


def _member_type_infos_cb(pkt, lst, cur, remain):
    """
    Returns a NRBFAdditionalInfo with the type selected.
    """
    # Get the next member of 'BinaryTypeEnums'
    index = len(lst) + (1 if cur is not None else 0)
    try:
        typeEnum = next(
            y
            for i, y in enumerate(
                x
                for x in pkt.BinaryTypeEnums
                if x
                not in [
                    # Some types are ignored (see table in [MS-NRBF] sect 2.3.1.2)
                    BinaryTypeEnum.String,
                    BinaryTypeEnum.Object,
                    BinaryTypeEnum.ObjectArray,
                    BinaryTypeEnum.StringArray,
                ]
            )
            if i >= index
        )
    except StopIteration:
        return None
    typeEnum = BinaryTypeEnum(typeEnum)
    # Return BinaryTypeEnum tainted with a pre-selected type.
    return functools.partial(
        NRBFAdditionalInfo,
        type=typeEnum,
    )


class NRBFMemberTypeInfo(Packet):
    fields_desc = [
        FieldListField(
            "BinaryTypeEnums",
            [],
            ByteEnumField("", 0, BinaryTypeEnum),
            count_from=lambda pkt: pkt.MemberCount,
        ),
        PacketListField(
            "AdditionalInfos",
            [],
            None,
            next_cls_cb=_member_type_infos_cb,
        ),
    ]


# [MS-NRBF] 2.3.2.5


class NRBFClassWithId(NRBFRecord):
    RecordTypeEnum = 1
    fields_desc = [
        NRBFRecord,
        LESignedIntField("ObjectId", 0),
        LESignedIntField("MetadataId", 0),
    ]


# [MS-NRBF] sect 2.5.2


class NRBFMemberPrimitiveUnTyped(Packet):
    __slots__ = ["type"]

    fields_desc = [
        NRBFValueWithCode.fields_desc[1],
    ]

    def __init__(self, _pkt=None, **kwargs):
        self.type = kwargs.pop("type", PrimitiveTypeEnum.Byte)
        assert isinstance(self.type, PrimitiveTypeEnum)
        super(NRBFMemberPrimitiveUnTyped, self).__init__(_pkt, **kwargs)

    def clone_with(self, *args, **kwargs):
        pkt = super(NRBFMemberPrimitiveUnTyped, self).clone_with(*args, **kwargs)
        pkt.type = self.type
        return pkt

    def copy(self):
        pkt = super(NRBFMemberPrimitiveUnTyped, self).copy()
        pkt.type = self.type
        return pkt

    @property
    def PrimitiveType(self):
        return self.type

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-NRBF] sect 2.3.2.1


class NRBFClassWithMembersAndTypes(NRBFRecord):
    RecordTypeEnum = 5
    fields_desc = [
        NRBFRecord,
        NRBFClassInfo,
        NRBFMemberTypeInfo,
        LESignedIntField("LibraryId", 0),
        _NRBFMembers,
    ]


# [MS-NRBF] sect 2.3.2.3


class NRBFSystemClassWithMembersAndTypes(NRBFRecord):
    RecordTypeEnum = 4
    fields_desc = [
        NRBFRecord,
        NRBFClassInfo,
        NRBFMemberTypeInfo,
        _NRBFMembers,
    ]


# [MS-NRBF] sect 2.3.2.4


class NRBFSystemClassWithMembers(NRBFRecord):
    RecordTypeEnum = 2
    fields_desc = [
        NRBFRecord,
        NRBFClassInfo,
        _NRBFMembers,
    ]


# [MS-NRBF] sect 2.4.2.1


class ArrayInfo(Packet):
    fields_desc = [LEIntField("ObjectId", 0), LEIntField("Length", None)]


# [MS-NRBF] sect 2.4.3.2


class NRBFArraySingleObject(NRBFRecord):
    RecordTypeEnum = 16
    Length = 1
    fields_desc = [
        NRBFRecord,
        ArrayInfo,
    ]


# [MS-NRBF] sect 2.4.3.3


def _values_singleprim_cb(pkt, lst, cur, remain):
    index = len(lst) + (1 if cur is not None else 0)
    if index >= pkt.Length:
        return None
    return functools.partial(
        NRBFMemberPrimitiveUnTyped,
        type=PrimitiveTypeEnum(pkt.PrimitiveTypeEnum),
    )


class NRBFArraySinglePrimitive(NRBFRecord):
    RecordTypeEnum = 15
    fields_desc = [
        NRBFRecord,
        ArrayInfo,
        ByteEnumField("PrimitiveTypeEnum", 0, PrimitiveTypeEnum),
        MultipleTypeField(
            [
                (
                    StrLenField("Values", [], length_from=lambda pkt: pkt.Length),
                    lambda pkt: pkt.PrimitiveTypeEnum == PrimitiveTypeEnum.Byte,
                )
            ],
            PacketListField(
                "Values",
                [],
                next_cls_cb=_values_singleprim_cb,
                max_count=1000,
            ),
        ),
    ]

    def post_build(self, p, pay):
        if self.Length is None:
            p = p[:5] + struct.pack("<I", len(self.Values)) + p[9:]
        return p + pay


# [MS-NRBF] sect 2.5.1


class NRBFMemberPrimitiveTyped(NRBFRecord):
    RecordTypeEnum = 9
    fields_desc = [
        NRBFRecord,
        NRBFValueWithCode,
    ]


# [MS-NRBF] sect 2.5.3


class NRBFMemberReference(NRBFRecord):
    RecordTypeEnum = 9
    fields_desc = [
        NRBFRecord,
        LEIntField("IdRef", 0),
    ]


# [MS-NRBF] sect 2.5.4


class NRBFObjectNull(NRBFRecord):
    RecordTypeEnum = 10
    fields_desc = [
        NRBFRecord,
    ]


# [MS-NRBF] sect 2.5.7


class NRBFBinaryObjectString(NRBFRecord):
    RecordTypeEnum = 6
    fields_desc = [
        NRBFRecord,
        LEIntField("ObjectId", 0),
        PacketField("Value", NRBFLengthPrefixedString(), NRBFLengthPrefixedString),
    ]


# [MS-NRBF] sect 2.6.1


class NRBFSerializationHeader(NRBFRecord):
    RecordTypeEnum = 0
    fields_desc = [
        NRBFRecord,
        LESignedIntField("RootID", 1),
        LESignedIntField("HeaderId", 0),
        LESignedIntField("MajorVersion", 1),
        LESignedIntField("MinorVersion", 0),
    ]


# [MS-NRBF] sect 2.6.2


class NRBFBinaryLibrary(NRBFRecord):
    RecordTypeEnum = 12
    fields_desc = [
        NRBFRecord,
        LESignedIntField("LibraryId", 0),
        PacketField(
            "LibraryName", NRBFLengthPrefixedString(), NRBFLengthPrefixedString
        ),
    ]


# [MS-NRBF] sect 2.6.3


class NRBFMessageEnd(NRBFRecord):
    RecordTypeEnum = 11
    fields_desc = [
        NRBFRecord,
    ]


# NRBF is a list of records


def _nrbf_records_cls(pkt, lst, cur, remain):
    # Detect end
    if isinstance(cur, NRBFMessageEnd):
        return None
    return NRBFRecord


class NRBF(Packet):
    # This is the same structure as what is returned by ysoserial.net
    fields_desc = [
        PacketListField(
            "records",
            [NRBFMessageEnd()],
            None,
            max_count=1000,
            next_cls_cb=_nrbf_records_cls,
        )
    ]

    def default_payload_class(self, _):
        return conf.padding_layer


# Automatons


class NRTP_Server(Automaton):
    """
    NRTP server to send a single payload.
    """

    pkt_cls = NRTPSingleMessageContent
    socketcls = StreamSocket

    def __init__(self, PAYLOAD, verb=True, *args, **kwargs):
        self.PAYLOAD = PAYLOAD
        self.verb = verb
        if "sock" not in kwargs:
            raise ValueError(
                "NMF_Server cannot be started directly ! Use NMF_Server.spawn"
            )
        Automaton.__init__(
            self,
            *args,
            **kwargs,
        )

    @ATMT.state(initial=1)
    def BEGIN(self):
        pass

    @ATMT.state(error=1)
    def FAILURE(self, error):
        return error

    @ATMT.receive_condition(BEGIN)
    def should_send_response(self, pkt):
        if NRTPSingleMessageContent in pkt:
            raise self.END().action_parameters(pkt)

    @ATMT.action(should_send_response)
    def send_response(self, pkt):
        self.send(NRTPSingleMessageContent(OperationType="Reply") / self.PAYLOAD)

    @ATMT.state(final=1)
    def END(self):
        pass


if __name__ == "__main__":
    interact(mydict=globals(), mybanner="Scapy [MS-NRTP] addon")



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = CAN Calibration Protocol (CCP)
# scapy.contrib.status = loads

import struct

from scapy.packet import Packet, bind_layers, bind_bottom_up
from scapy.fields import XIntField, FlagsField, ByteEnumField, \
    ThreeBytesField, XBitField, ShortField, IntField, XShortField, \
    ByteField, XByteField, StrFixedLenField, LEShortField
from scapy.layers.can import CAN


class CCP(CAN):
    name = 'CAN Calibration Protocol'
    fields_desc = [
        FlagsField('flags', 0, 3, ['error',
                                   'remote_transmission_request',
                                   'extended']),
        XBitField('identifier', 0, 29),
        ByteField('length', 8),
        ThreeBytesField('reserved', 0),
    ]

    def extract_padding(self, p):
        return p, None


class CRO(Packet):
    commands = {
        0x01: "CONNECT",
        0x1B: "GET_CCP_VERSION",
        0x17: "EXCHANGE_ID",
        0x12: "GET_SEED",
        0x13: "UNLOCK",
        0x02: "SET_MTA",
        0x03: "DNLOAD",
        0x23: "DNLOAD_6",
        0x04: "UPLOAD",
        0x0F: "SHORT_UP",
        0x11: "SELECT_CAL_PAGE",
        0x14: "GET_DAQ_SIZE",
        0x15: "SET_DAQ_PTR",
        0x16: "WRITE_DAQ",
        0x06: "START_STOP",
        0x07: "DISCONNECT",
        0x0C: "SET_S_STATUS",
        0x0D: "GET_S_STATUS",
        0x0E: "BUILD_CHKSUM",
        0x10: "CLEAR_MEMORY",
        0x18: "PROGRAM",
        0x22: "PROGRAM_6",
        0x19: "MOVE",
        0x05: "TEST",
        0x09: "GET_ACTIVE_CAL_PAGE",
        0x08: "START_STOP_ALL",
        0x20: "DIAG_SERVICE",
        0x21: "ACTION_SERVICE"
    }
    name = 'Command Receive Object'
    fields_desc = [
        ByteEnumField('cmd', 0x01, commands),
        ByteField('ctr', 0)
    ]

    def hashret(self):
        return struct.pack('B', self.ctr)


# ##### CROs ######

class CONNECT(Packet):
    fields_desc = [
        LEShortField('station_address', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4),
    ]


bind_layers(CRO, CONNECT, cmd=0x01)


class GET_CCP_VERSION(Packet):
    fields_desc = [
        XByteField('main_protocol_version', 0),
        XByteField('release_version', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4)
    ]


bind_layers(CRO, GET_CCP_VERSION, cmd=0x1B)


class EXCHANGE_ID(Packet):
    fields_desc = [
        StrFixedLenField('ccp_master_device_id', b'\x00' * 6, length=6)
    ]


bind_layers(CRO, EXCHANGE_ID, cmd=0x17)


class GET_SEED(Packet):
    fields_desc = [
        XByteField('resource', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 5, length=5)
    ]


bind_layers(CRO, GET_SEED, cmd=0x12)


class UNLOCK(Packet):
    fields_desc = [
        StrFixedLenField('key', b'\x00' * 6, length=6)
    ]


bind_layers(CRO, UNLOCK, cmd=0x13)


class SET_MTA(Packet):
    fields_desc = [
        XByteField('mta_num', 0),
        XByteField('address_extension', 0),
        XIntField('address', 0),
    ]


bind_layers(CRO, SET_MTA, cmd=0x02)


class DNLOAD(Packet):
    fields_desc = [
        XByteField('size', 0),
        StrFixedLenField('data', b'\x00' * 5, length=5)
    ]


bind_layers(CRO, DNLOAD, cmd=0x03)


class DNLOAD_6(Packet):
    fields_desc = [
        StrFixedLenField('data', b'\x00' * 6, length=6)
    ]


bind_layers(CRO, DNLOAD_6, cmd=0x23)


class UPLOAD(Packet):
    fields_desc = [
        XByteField('size', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 5, length=5)
    ]


bind_layers(CRO, UPLOAD, cmd=0x04)


class SHORT_UP(Packet):
    fields_desc = [
        XByteField('size', 0),
        XByteField('address_extension', 0),
        XIntField('address', 0),
    ]


bind_layers(CRO, SHORT_UP, cmd=0x0F)


class SELECT_CAL_PAGE(Packet):
    fields_desc = [
        StrFixedLenField('ccp_reserved', b'\xff' * 6, length=6)
    ]


bind_layers(CRO, SELECT_CAL_PAGE, cmd=0x11)


class GET_DAQ_SIZE(Packet):
    fields_desc = [
        XByteField('DAQ_num', 0),
        XByteField('ccp_reserved', 0),
        XIntField('DTO_identifier', 0),
    ]


bind_layers(CRO, GET_DAQ_SIZE, cmd=0x14)


class SET_DAQ_PTR(Packet):
    fields_desc = [
        XByteField('DAQ_num', 0),
        XByteField('ODT_num', 0),
        XByteField('ODT_element', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 3, length=3)
    ]


bind_layers(CRO, SET_DAQ_PTR, cmd=0x15)


class WRITE_DAQ(Packet):
    fields_desc = [
        XByteField('DAQ_size', 0),
        XByteField('address_extension', 0),
        XIntField('address', 0),
    ]


bind_layers(CRO, WRITE_DAQ, cmd=0x16)


class START_STOP(Packet):
    fields_desc = [
        XByteField('mode', 0),
        XByteField('DAQ_num', 0),
        XByteField('ODT_num', 0),
        XByteField('event_channel', 0),
        XShortField('transmission_rate', 0),
    ]


bind_layers(CRO, START_STOP, cmd=0x06)


class DISCONNECT(Packet):
    fields_desc = [
        ByteEnumField('type', 0, {0: "temporary", 1: "end_of_session"}),
        StrFixedLenField('ccp_reserved0', b'\xff' * 1, length=1),
        LEShortField('station_address', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 2, length=2)
    ]


bind_layers(CRO, DISCONNECT, cmd=0x07)


class SET_S_STATUS(Packet):
    name = "Set Session Status"
    fields_desc = [
        FlagsField("session_status", 0, 8, ["CAL", "DAQ", "RESUME", "RES0",
                                            "RES1", "RES2", "STORE", "RUN"]),
        StrFixedLenField('ccp_reserved', b'\xff' * 5, length=5)
    ]


bind_layers(CRO, SET_S_STATUS, cmd=0x0C)


class GET_S_STATUS(Packet):
    fields_desc = [
        StrFixedLenField('ccp_reserved', b'\xff' * 6, length=6)
    ]


bind_layers(CRO, GET_S_STATUS, cmd=0x0D)


class BUILD_CHKSUM(Packet):
    fields_desc = [
        IntField('size', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 2, length=2)
    ]


bind_layers(CRO, BUILD_CHKSUM, cmd=0x0E)


class CLEAR_MEMORY(Packet):
    fields_desc = [
        IntField('size', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 2, length=2)
    ]


bind_layers(CRO, CLEAR_MEMORY, cmd=0x10)


class PROGRAM(Packet):
    fields_desc = [
        XByteField('size', 0),
        StrFixedLenField('data', b'\x00' * 0,
                         length_from=lambda pkt: pkt.size),
        StrFixedLenField('ccp_reserved', b'\xff' * 5,
                         length_from=lambda pkt: 5 - pkt.size)
    ]


bind_layers(CRO, PROGRAM, cmd=0x18)


class PROGRAM_6(Packet):
    fields_desc = [
        StrFixedLenField('data', b'\x00' * 6, length=6)
    ]


bind_layers(CRO, PROGRAM_6, cmd=0x22)


class MOVE(Packet):
    fields_desc = [
        IntField('size', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 2, length=2)
    ]


bind_layers(CRO, MOVE, cmd=0x19)


class TEST(Packet):
    fields_desc = [
        LEShortField('station_address', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4)
    ]


bind_layers(CRO, TEST, cmd=0x05)


class GET_ACTIVE_CAL_PAGE(Packet):
    fields_desc = [
        StrFixedLenField('ccp_reserved', b'\xff' * 6, length=6)
    ]


bind_layers(CRO, GET_ACTIVE_CAL_PAGE, cmd=0x09)


class START_STOP_ALL(Packet):
    fields_desc = [
        ByteEnumField('type', 0, {0: "stop", 1: "start"}),
        StrFixedLenField('ccp_reserved', b'\xff' * 5, length=5)

    ]


bind_layers(CRO, START_STOP_ALL, cmd=0x08)


class DIAG_SERVICE(Packet):
    fields_desc = [
        ShortField('diag_service', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4)
    ]


bind_layers(CRO, DIAG_SERVICE, cmd=0x20)


class ACTION_SERVICE(Packet):
    fields_desc = [
        ShortField('action_service', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4)
    ]


bind_layers(CRO, ACTION_SERVICE, cmd=0x21)


# ##### DTOs ######

class DEFAULT_DTO(Packet):
    fields_desc = [
        StrFixedLenField('load', b'\xff' * 5, length=5),
    ]


class GET_CCP_VERSION_DTO(Packet):
    fields_desc = [
        XByteField('main_protocol_version', 0),
        XByteField('release_version', 0),
        StrFixedLenField('ccp_reserved', b'\x00' * 3, length=3)
    ]


class EXCHANGE_ID_DTO(Packet):
    fields_desc = [
        ByteField('slave_device_ID_length', 0),
        ByteField('data_type_qualifier', 0),
        ByteField('resource_availability_mask', 0),
        ByteField('resource_protection_mask', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 1, length=1),
    ]


class GET_SEED_DTO(Packet):
    fields_desc = [
        XByteField('protection_status', 0),
        StrFixedLenField('seed', b'\x00' * 4, length=4)
    ]


class UNLOCK_DTO(Packet):
    fields_desc = [
        ByteField('privilege_status', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 4, length=4),
    ]


class DNLOAD_DTO(Packet):
    fields_desc = [
        XByteField('MTA0_extension', 0),
        XIntField('MTA0_address', 0)
    ]


class DNLOAD_6_DTO(Packet):
    fields_desc = [
        XByteField('MTA0_extension', 0),
        XIntField('MTA0_address', 0)
    ]


class UPLOAD_DTO(Packet):
    fields_desc = [
        StrFixedLenField('data', b'\x00' * 5, length=5)
    ]


class SHORT_UP_DTO(Packet):
    fields_desc = [
        StrFixedLenField('data', b'\x00' * 5, length=5)
    ]


class GET_DAQ_SIZE_DTO(Packet):
    fields_desc = [
        XByteField('DAQ_list_size', 0),
        XByteField('first_pid', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 3, length=3)
    ]


class GET_S_STATUS_DTO(Packet):
    fields_desc = [
        FlagsField("session_status", 0, 8, ["CAL", "DAQ", "RESUME", "RES0",
                                            "RES1", "RES2", "STORE", "RUN"]),
        ByteField('information_qualifier', 0),
        StrFixedLenField('information', b'\x00' * 3, length=3)
    ]


class BUILD_CHKSUM_DTO(Packet):
    fields_desc = [
        ByteField('checksum_size', 0),
        StrFixedLenField('checksum_data', b'\x00' * 4,
                         length_from=lambda pkt: pkt.checksum_size),
        StrFixedLenField('ccp_reserved', b'\xff' * 0,
                         length_from=lambda pkt: 4 - pkt.checksum_size)
    ]


class PROGRAM_DTO(Packet):
    fields_desc = [
        ByteField('MTA0_extension', 0),
        XIntField('MTA0_address', 0)
    ]


class PROGRAM_6_DTO(Packet):
    fields_desc = [
        ByteField('MTA0_extension', 0),
        XIntField('MTA0_address', 0)
    ]


class GET_ACTIVE_CAL_PAGE_DTO(Packet):
    fields_desc = [
        XByteField('address_extension', 0),
        XIntField('address', 0)
    ]


class DIAG_SERVICE_DTO(Packet):
    fields_desc = [
        ByteField('data_length', 0),
        ByteField('data_type', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 3, length=3)
    ]


class ACTION_SERVICE_DTO(Packet):
    fields_desc = [
        ByteField('data_length', 0),
        ByteField('data_type', 0),
        StrFixedLenField('ccp_reserved', b'\xff' * 3, length=3)
    ]


class DTO(Packet):
    __slots__ = Packet.__slots__ + ["payload_cls"]

    return_codes = {
        0x00: "acknowledge / no error",
        0x01: "DAQ processor overload",
        0x10: "command processor busy",
        0x11: "DAQ processor busy",
        0x12: "internal timeout",
        0x18: "key request",
        0x19: "session status request",
        0x20: "cold start request",
        0x21: "cal. data init. request",
        0x22: "DAQ list init. request",
        0x23: "code update request",
        0x30: "unknown command",
        0x31: "command syntax",
        0x32: "parameter(s) out of range",
        0x33: "access denied",
        0x34: "overload",
        0x35: "access locked",
        0x36: "resource/function not available"
    }
    fields_desc = [
        XByteField("packet_id", 0xff),
        ByteEnumField('return_code', 0x00, return_codes),
        ByteField('ctr', 0)
    ]

    def __init__(self, *args, **kwargs):
        self.payload_cls = DEFAULT_DTO
        if "payload_cls" in kwargs:
            self.payload_cls = kwargs["payload_cls"]
            del kwargs["payload_cls"]
        Packet.__init__(self, *args, **kwargs)

    def __eq__(self, other):
        return super(DTO, self).__eq__(other) and \
            self.payload_cls == other.payload_cls

    def guess_payload_class(self, payload):
        return self.payload_cls

    @staticmethod
    def get_dto_cls(cmd):
        try:
            return {
                0x03: DNLOAD_DTO,
                0x04: UPLOAD_DTO,
                0x09: GET_ACTIVE_CAL_PAGE_DTO,
                0x0D: GET_S_STATUS_DTO,
                0x0E: BUILD_CHKSUM_DTO,
                0x0F: SHORT_UP_DTO,
                0x12: GET_SEED_DTO,
                0x13: UNLOCK_DTO,
                0x14: GET_DAQ_SIZE_DTO,
                0x17: EXCHANGE_ID_DTO,
                0x18: PROGRAM_DTO,
                0x1B: GET_CCP_VERSION_DTO,
                0x20: DIAG_SERVICE_DTO,
                0x21: ACTION_SERVICE_DTO,
                0x22: PROGRAM_6_DTO,
                0x23: DNLOAD_6_DTO
            }[cmd]
        except KeyError:
            return DEFAULT_DTO

    def answers(self, other):
        """In CCP, the payload of a DTO packet is dependent on the cmd field
        of a corresponding CRO packet. Two packets correspond, if there
        ctr field is equal. If answers detect the corresponding CRO, it will
        interpret the payload of a DTO with the correct class. In CCP, there is
        no other way, to determine the class of a DTO payload. Since answers is
        called on sr and sr1, this modification of the original answers
        implementation will give a better user experience. """
        if not hasattr(other, "ctr"):
            return 0
        if self.ctr != other.ctr:
            return 0
        if not hasattr(other, "cmd"):
            return 0

        new_pl_cls = self.get_dto_cls(other.cmd)
        if self.payload_cls != new_pl_cls and \
                self.payload_cls == DEFAULT_DTO:
            data = bytes(self.load)
            self.remove_payload()
            self.add_payload(new_pl_cls(data))
            self.payload_cls = new_pl_cls
        return 1

    def hashret(self):
        return struct.pack('B', self.ctr)


bind_bottom_up(CCP, DTO)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Michael Farrell <micolous+git@gmail.com>

"""
nRF sniffer

Firmware and documentation related to this module is available at:
https://www.nordicsemi.com/Software-and-Tools/Development-Tools/nRF-Sniffer
https://github.com/adafruit/Adafruit_BLESniffer_Python
https://github.com/wireshark/wireshark/blob/master/epan/dissectors/packet-nordic_ble.c
"""

# scapy.contrib.description = nRF sniffer
# scapy.contrib.status = works

import struct

from scapy.config import conf
from scapy.data import DLT_NORDIC_BLE
from scapy.fields import (
    BitEnumField,
    BitField,
    ByteEnumField,
    ByteField,
    LEIntField,
    LEShortField,
    LenField,
    ScalingField,
)
from scapy.layers.bluetooth4LE import BTLE
from scapy.packet import Packet, bind_layers


# nRF Sniffer v2


class NRFS2_Packet(Packet):
    """
    nRF Sniffer v2 Packet
    """

    fields_desc = [
        LenField("len", None, fmt="<H", adjust=lambda x: x + 6),
        ByteField("version", 2),
        LEShortField("counter", None),
        ByteEnumField(
            "type",
            None,
            {
                0x00: "req_follow",
                0x01: "event_follow",
                0x02: "event_device",  # missing from spreadsheet
                0x03: "req_single_packet",  # missing from spreadsheet
                0x04: "resp_single_packet",  # missing from spreadsheet
                0x05: "event_connect",
                0x06: "event_packet",
                0x07: "req_scan_cont",
                0x09: "event_disconnect",
                0x0A: "event_error",  # missing from spreadsheet
                0x0B: "event_empty_data_packet",  # missing from spreadsheet
                0x0C: "set_temporary_key",
                0x0D: "ping_req",
                0x0E: "ping_resp",
                0x0F: "test_command_id",  # missing from spreadsheet
                0x10: "test_result_id",  # missing from spreadsheet
                0x11: "uart_test_start",  # missing from spreadsheet
                0x12: "uart_dummy_packet",  # missing from spreadsheet
                0x13: "switch_baud_rate_req",  # not implemented in FW
                0x14: "switch_baud_rate_resp",  # not implemented in FW
                0x15: "uart_out_start",  # missing from spreadsheet
                0x16: "uart_out_stop",  # missing from spreadsheet
                0x17: "set_adv_channel_hop_seq",
                0xFE: "go_idle",  # not implemented in FW
            },
        ),
    ]

    def answer(self, other):
        if not isinstance(other, NRFS2_Packet):
            return False

        return (
            (self.type == 0x01 and other.type == 0x00) or
            (self.type == 0x0E and other.type == 0x0D) or
            (self.type == 0x14 and other.type == 0x13)
        )

    def post_build(self, p, pay):
        if self.hdr_len is None:
            p = p[:1] + struct.pack("!B", len(p)) + p[2:]
        return p + pay


class NRF2_Ping_Request(Packet):
    name = "Ping request"


class NRF2_Ping_Response(Packet):
    name = "Ping response"
    fields_desc = [
        LEShortField("version", None),
    ]


class NRF2_Packet_Event(Packet):
    name = "Packet event (device variant)"
    fields_desc = [
        ByteField("header_len", 10),
        # Flags (1 byte)
        BitField("reserved", 0, 1),
        BitEnumField("phy", None, 3, {0: "le-1m", 1: "le-2m", 2: "le-coded"}),
        BitField("mic", None, 1),
        BitField("encrypted", None, 1),
        BitField("direction", None, 1),
        BitField("crc_ok", 1, 1),
        ByteField("rf_channel", 0),
        ScalingField("rssi", -256, unit="dBm", fmt="b"),
        LEShortField("event_counter", 0),
        LEIntField("delta_time", 0),  # microseconds
    ]


bind_layers(NRFS2_Packet, NRF2_Ping_Request, type=0xD)
bind_layers(NRFS2_Packet, NRF2_Ping_Response, type=0xE)
bind_layers(NRFS2_Packet, NRF2_Packet_Event, type=0x6)

bind_layers(NRF2_Packet_Event, BTLE)

# Wire transport


class NRFS2_PCAP(Packet):
    """
    PCAP headers for DLT_NORDIC_BLE.

    Nordic's capture scripts either stick the COM port number (yep!) or a
    random number at the start of every packet.

    https://github.com/wireshark/wireshark/blob/master/epan/dissectors/packet-nordic_ble.c

    The only "rule" is that we can't start packets with ``BE EF``, otherwise
    it becomes a "0.9.7" packet. So we just set "0" here.
    """

    name = "nRF Sniffer PCAP header"
    fields_desc = [
        ByteField("board_id", 0),
    ]


bind_layers(NRFS2_PCAP, NRFS2_Packet)
conf.l2types.register(DLT_NORDIC_BLE, NRFS2_PCAP)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
#  Copyright (C) 2006 Nicolas Bareil <nicolas.bareil <at> eads.net>
#                     EADS/CRC security team

# scapy.contrib.description = Skinny Call Control Protocol (SCCP)
# scapy.contrib.status = loads

"""
Skinny Call Control Protocol (SCCP) extension
"""

import time
import struct

from scapy.packet import Packet, bind_layers
from scapy.fields import FlagsField, IPField, LEIntEnumField, LEIntField, \
    StrFixedLenField
from scapy.layers.inet import TCP
from scapy.volatile import RandShort
from scapy.config import conf

#####################################################################
# Helpers and constants
#####################################################################

skinny_messages_cls = {
    # Station -> Callmanager
    0x0000: "SkinnyMessageKeepAlive",
    0x0001: "SkinnyMessageRegister",
    0x0002: "SkinnyMessageIpPort",
    0x0003: "SkinnyMessageKeypadButton",
    0x0004: "SkinnyMessageEnblocCall",
    0x0005: "SkinnyMessageStimulus",
    0x0006: "SkinnyMessageOffHook",
    0x0007: "SkinnyMessageOnHook",
    0x0008: "SkinnyMessageHookFlash",
    0x0009: "SkinnyMessageForwardStatReq",
    0x000A: "SkinnyMessageSpeedDialStatReq",
    0x000B: "SkinnyMessageLineStatReq",
    0x000C: "SkinnyMessageConfigStatReq",
    0x000D: "SkinnyMessageTimeDateReq",
    0x000E: "SkinnyMessageButtonTemplateReq",
    0x000F: "SkinnyMessageVersionReq",
    0x0010: "SkinnyMessageCapabilitiesRes",
    0x0011: "SkinnyMessageMediaPortList",
    0x0012: "SkinnyMessageServerReq",
    0x0020: "SkinnyMessageAlarm",
    0x0021: "SkinnyMessageMulticastMediaReceptionAck",
    0x0022: "SkinnyMessageOpenReceiveChannelAck",
    0x0023: "SkinnyMessageConnectionStatisticsRes",
    0x0024: "SkinnyMessageOffHookWithCgpn",
    0x0025: "SkinnyMessageSoftKeySetReq",
    0x0026: "SkinnyMessageSoftKeyEvent",
    0x0027: "SkinnyMessageUnregister",
    0x0028: "SkinnyMessageSoftKeyTemplateReq",
    0x0029: "SkinnyMessageRegisterTokenReq",
    0x002A: "SkinnyMessageMediaTransmissionFailure",
    0x002B: "SkinnyMessageHeadsetStatus",
    0x002C: "SkinnyMessageMediaResourceNotification",
    0x002D: "SkinnyMessageRegisterAvailableLines",
    0x002E: "SkinnyMessageDeviceToUserData",
    0x002F: "SkinnyMessageDeviceToUserDataResponse",
    0x0030: "SkinnyMessageUpdateCapabilities",
    0x0031: "SkinnyMessageOpenMultiMediaReceiveChannelAck",
    0x0032: "SkinnyMessageClearConference",
    0x0033: "SkinnyMessageServiceURLStatReq",
    0x0034: "SkinnyMessageFeatureStatReq",
    0x0035: "SkinnyMessageCreateConferenceRes",
    0x0036: "SkinnyMessageDeleteConferenceRes",
    0x0037: "SkinnyMessageModifyConferenceRes",
    0x0038: "SkinnyMessageAddParticipantRes",
    0x0039: "SkinnyMessageAuditConferenceRes",
    0x0040: "SkinnyMessageAuditParticipantRes",
    0x0041: "SkinnyMessageDeviceToUserDataVersion1",
    # Callmanager -> Station */
    0x0081: "SkinnyMessageRegisterAck",
    0x0082: "SkinnyMessageStartTone",
    0x0083: "SkinnyMessageStopTone",
    0x0085: "SkinnyMessageSetRinger",
    0x0086: "SkinnyMessageSetLamp",
    0x0087: "SkinnyMessageSetHkFDetect",
    0x0088: "SkinnyMessageSpeakerMode",
    0x0089: "SkinnyMessageSetMicroMode",
    0x008A: "SkinnyMessageStartMediaTransmission",
    0x008B: "SkinnyMessageStopMediaTransmission",
    0x008C: "SkinnyMessageStartMediaReception",
    0x008D: "SkinnyMessageStopMediaReception",
    0x008F: "SkinnyMessageCallInfo",
    0x0090: "SkinnyMessageForwardStat",
    0x0091: "SkinnyMessageSpeedDialStat",
    0x0092: "SkinnyMessageLineStat",
    0x0093: "SkinnyMessageConfigStat",
    0x0094: "SkinnyMessageTimeDate",
    0x0095: "SkinnyMessageStartSessionTransmission",
    0x0096: "SkinnyMessageStopSessionTransmission",
    0x0097: "SkinnyMessageButtonTemplate",
    0x0098: "SkinnyMessageVersion",
    0x0099: "SkinnyMessageDisplayText",
    0x009A: "SkinnyMessageClearDisplay",
    0x009B: "SkinnyMessageCapabilitiesReq",
    0x009C: "SkinnyMessageEnunciatorCommand",
    0x009D: "SkinnyMessageRegisterReject",
    0x009E: "SkinnyMessageServerRes",
    0x009F: "SkinnyMessageReset",
    0x0100: "SkinnyMessageKeepAliveAck",
    0x0101: "SkinnyMessageStartMulticastMediaReception",
    0x0102: "SkinnyMessageStartMulticastMediaTransmission",
    0x0103: "SkinnyMessageStopMulticastMediaReception",
    0x0104: "SkinnyMessageStopMulticastMediaTransmission",
    0x0105: "SkinnyMessageOpenReceiveChannel",
    0x0106: "SkinnyMessageCloseReceiveChannel",
    0x0107: "SkinnyMessageConnectionStatisticsReq",
    0x0108: "SkinnyMessageSoftKeyTemplateRes",
    0x0109: "SkinnyMessageSoftKeySetRes",
    0x0110: "SkinnyMessageStationSelectSoftKeysMessage",
    0x0111: "SkinnyMessageCallState",
    0x0112: "SkinnyMessagePromptStatus",
    0x0113: "SkinnyMessageClearPromptStatus",
    0x0114: "SkinnyMessageDisplayNotify",
    0x0115: "SkinnyMessageClearNotify",
    0x0116: "SkinnyMessageCallPlane",
    0x0117: "SkinnyMessageCallPlane",
    0x0118: "SkinnyMessageUnregisterAck",
    0x0119: "SkinnyMessageBackSpaceReq",
    0x011A: "SkinnyMessageRegisterTokenAck",
    0x011B: "SkinnyMessageRegisterTokenReject",
    0x0042: "SkinnyMessageDeviceToUserDataResponseVersion1",
    0x011C: "SkinnyMessageStartMediaFailureDetection",
    0x011D: "SkinnyMessageDialedNumber",
    0x011E: "SkinnyMessageUserToDeviceData",
    0x011F: "SkinnyMessageFeatureStat",
    0x0120: "SkinnyMessageDisplayPriNotify",
    0x0121: "SkinnyMessageClearPriNotify",
    0x0122: "SkinnyMessageStartAnnouncement",
    0x0123: "SkinnyMessageStopAnnouncement",
    0x0124: "SkinnyMessageAnnouncementFinish",
    0x0127: "SkinnyMessageNotifyDtmfTone",
    0x0128: "SkinnyMessageSendDtmfTone",
    0x0129: "SkinnyMessageSubscribeDtmfPayloadReq",
    0x012A: "SkinnyMessageSubscribeDtmfPayloadRes",
    0x012B: "SkinnyMessageSubscribeDtmfPayloadErr",
    0x012C: "SkinnyMessageUnSubscribeDtmfPayloadReq",
    0x012D: "SkinnyMessageUnSubscribeDtmfPayloadRes",
    0x012E: "SkinnyMessageUnSubscribeDtmfPayloadErr",
    0x012F: "SkinnyMessageServiceURLStat",
    0x0130: "SkinnyMessageCallSelectStat",
    0x0131: "SkinnyMessageOpenMultiMediaChannel",
    0x0132: "SkinnyMessageStartMultiMediaTransmission",
    0x0133: "SkinnyMessageStopMultiMediaTransmission",
    0x0134: "SkinnyMessageMiscellaneousCommand",
    0x0135: "SkinnyMessageFlowControlCommand",
    0x0136: "SkinnyMessageCloseMultiMediaReceiveChannel",
    0x0137: "SkinnyMessageCreateConferenceReq",
    0x0138: "SkinnyMessageDeleteConferenceReq",
    0x0139: "SkinnyMessageModifyConferenceReq",
    0x013A: "SkinnyMessageAddParticipantReq",
    0x013B: "SkinnyMessageDropParticipantReq",
    0x013C: "SkinnyMessageAuditConferenceReq",
    0x013D: "SkinnyMessageAuditParticipantReq",
    0x013F: "SkinnyMessageUserToDeviceDataVersion1",
}

skinny_callstates = {
    0x1: "Off Hook",
    0x2: "On Hook",
    0x3: "Ring out",
    0xc: "Proceeding",
}


skinny_ring_type = {
    0x1: "Ring off"
}

skinny_speaker_modes = {
    0x1: "Speaker on",
    0x2: "Speaker off"
}

skinny_lamp_mode = {
    0x1: "Off (?)",
    0x2: "On",
}

skinny_stimulus = {
    0x9: "Line"
}


############
#  Fields  #
############

class SkinnyDateTimeField(StrFixedLenField):
    def __init__(self, name, default):
        StrFixedLenField.__init__(self, name, default, 32)

    def m2i(self, pkt, s):
        year, month, dow, day, hour, min, sec, millisecond = struct.unpack('<8I', s)  # noqa: E501
        return (year, month, day, hour, min, sec)

    def i2m(self, pkt, val):
        if isinstance(val, str):
            val = self.h2i(pkt, val)
        tmp_lst = val[:2] + (0,) + val[2:7] + (0,)
        return struct.pack('<8I', *tmp_lst)

    def i2h(self, pkt, x):
        if isinstance(x, str):
            return x
        else:
            return time.ctime(time.mktime(x + (0, 0, 0)))

    def i2repr(self, pkt, x):
        return self.i2h(pkt, x)

    def h2i(self, pkt, s):
        t = ()
        if isinstance(s, str):
            t = time.strptime(s)
            t = t[:2] + t[2:-3]
        else:
            if not s:
                y, m, d, h, min, sec, rest, rest, rest = time.gmtime(time.time())  # noqa: E501
                t = (y, m, d, h, min, sec)
            else:
                t = s
        return t


###########################
#  Packet abstract class  #
###########################

class SkinnyMessageGeneric(Packet):
    name = 'Generic message'


class SkinnyMessageKeepAlive(Packet):
    name = 'keep alive'


class SkinnyMessageKeepAliveAck(Packet):
    name = 'keep alive ack'


class SkinnyMessageOffHook(Packet):
    name = 'Off Hook'
    fields_desc = [LEIntField("unknown1", 0),
                   LEIntField("unknown2", 0), ]


class SkinnyMessageOnHook(SkinnyMessageOffHook):
    name = 'On Hook'


class SkinnyMessageCallState(Packet):
    name = 'Skinny Call state message'
    fields_desc = [LEIntEnumField("state", 1, skinny_callstates),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0),
                   LEIntField("unknown1", 4),
                   LEIntField("unknown2", 0),
                   LEIntField("unknown3", 0)]


class SkinnyMessageSoftKeyEvent(Packet):
    name = 'Soft Key Event'
    fields_desc = [LEIntField("key", 0),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageSetRinger(Packet):
    name = 'Ring message'
    fields_desc = [LEIntEnumField("ring", 0x1, skinny_ring_type),
                   LEIntField("unknown1", 0),
                   LEIntField("unknown2", 0),
                   LEIntField("unknown3", 0)]


_skinny_tones = {
    0x21: 'Inside dial tone',
    0x22: 'xxx',
    0x23: 'xxx',
    0x24: 'Alerting tone',
    0x25: 'Reorder Tone'
}


class SkinnyMessageStartTone(Packet):
    name = 'Start tone'
    fields_desc = [LEIntEnumField("tone", 0x21, _skinny_tones),
                   LEIntField("unknown1", 0),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageStopTone(SkinnyMessageGeneric):
    name = 'stop tone'
    fields_desc = [LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageSpeakerMode(Packet):
    name = 'Speaker mdoe'
    fields_desc = [LEIntEnumField("ring", 0x1, skinny_speaker_modes)]


class SkinnyMessageSetLamp(Packet):
    name = 'Lamp message (light of the phone)'
    fields_desc = [LEIntEnumField("stimulus", 0x5, skinny_stimulus),
                   LEIntField("instance", 1),
                   LEIntEnumField("mode", 2, skinny_lamp_mode)]


class SkinnyMessageStationSelectSoftKeysMessage(Packet):
    name = 'Station Select Soft Keys Message'
    fields_desc = [LEIntField("instance", 1),
                   LEIntField("callid", 0),
                   LEIntField("set", 0),
                   LEIntField("map", 0xffff)]


class SkinnyMessagePromptStatus(Packet):
    name = 'Prompt status'
    fields_desc = [LEIntField("timeout", 0),
                   StrFixedLenField("text", b"\0" * 32, 32),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageCallPlane(Packet):
    name = 'Activate/Deactivate Call Plane Message'
    fields_desc = [LEIntField("instance", 1)]


class SkinnyMessageTimeDate(Packet):
    name = 'Setting date and time'
    fields_desc = [SkinnyDateTimeField("settime", None),
                   LEIntField("timestamp", 0)]


class SkinnyMessageClearPromptStatus(Packet):
    name = 'clear prompt status'
    fields_desc = [LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageKeypadButton(Packet):
    name = 'keypad button'
    fields_desc = [LEIntField("key", 0),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0)]


class SkinnyMessageDialedNumber(Packet):
    name = 'dialed number'
    fields_desc = [StrFixedLenField("number", "1337", 24),
                   LEIntField("instance", 1),
                   LEIntField("callid", 0)]


_skinny_message_callinfo_restrictions = ['CallerName', 'CallerNumber', 'CalledName', 'CalledNumber', 'OriginalCalledName', 'OriginalCalledNumber', 'LastRedirectName', 'LastRedirectNumber'] + ['Bit%d' % i for i in range(8, 15)]  # noqa: E501


class SkinnyMessageCallInfo(Packet):
    name = 'call information'
    fields_desc = [StrFixedLenField("callername", "Jean Valjean", 40),
                   StrFixedLenField("callernum", "1337", 24),
                   StrFixedLenField("calledname", "Causette", 40),
                   StrFixedLenField("callednum", "1034", 24),
                   LEIntField("lineinstance", 1),
                   LEIntField("callid", 0),
                   StrFixedLenField("originalcalledname", "Causette", 40),
                   StrFixedLenField("originalcallednum", "1034", 24),
                   StrFixedLenField("lastredirectingname", "Causette", 40),
                   StrFixedLenField("lastredirectingnum", "1034", 24),
                   LEIntField("originalredirectreason", 0),
                   LEIntField("lastredirectreason", 0),
                   StrFixedLenField('voicemailboxG', b'\0' * 24, 24),
                   StrFixedLenField('voicemailboxD', b'\0' * 24, 24),
                   StrFixedLenField('originalvoicemailboxD', b'\0' * 24, 24),
                   StrFixedLenField('lastvoicemailboxD', b'\0' * 24, 24),
                   LEIntField('security', 0),
                   FlagsField('restriction', 0, 16, _skinny_message_callinfo_restrictions),  # noqa: E501
                   LEIntField('unknown', 0)]


class SkinnyRateField(LEIntField):
    def i2repr(self, pkt, x):
        if x is None:
            x = 0
        return '%d ms/pkt' % x


_skinny_codecs = {
    0x0: 'xxx',
    0x1: 'xxx',
    0x2: 'xxx',
    0x3: 'xxx',
    0x4: 'G711 ulaw 64k'
}

_skinny_echo = {
    0x0: 'echo cancellation off',
    0x1: 'echo cancellation on'
}


class SkinnyMessageOpenReceiveChannel(Packet):
    name = 'open receive channel'
    fields_desc = [LEIntField('conference', 0),
                   LEIntField('passthru', 0),
                   SkinnyRateField('rate', 20),
                   LEIntEnumField('codec', 4, _skinny_codecs),
                   LEIntEnumField('echo', 0, _skinny_echo),
                   LEIntField('unknown1', 0),
                   LEIntField('callid', 0)]

    def guess_payload_class(self, p):
        return conf.padding_layer


_skinny_receive_channel_status = {
    0x0: 'ok',
    0x1: 'ko'
}


class SkinnyMessageOpenReceiveChannelAck(Packet):
    name = 'open receive channel'
    fields_desc = [LEIntEnumField('status', 0, _skinny_receive_channel_status),
                   IPField('remote', '0.0.0.0'),
                   LEIntField('port', RandShort()),
                   LEIntField('passthru', 0),
                   LEIntField('callid', 0)]


_skinny_silence = {
    0x0: 'silence suppression off',
    0x1: 'silence suppression on',
}


class SkinnyFramePerPacketField(LEIntField):
    def i2repr(self, pkt, x):
        if x is None:
            x = 0
        return '%d frames/pkt' % x


class SkinnyMessageStartMediaTransmission(Packet):
    name = 'start multimedia transmission'
    fields_desc = [LEIntField('conference', 0),
                   LEIntField('passthru', 0),
                   IPField('remote', '0.0.0.0'),
                   LEIntField('port', RandShort()),
                   SkinnyRateField('rate', 20),
                   LEIntEnumField('codec', 4, _skinny_codecs),
                   LEIntField('precedence', 200),
                   LEIntEnumField('silence', 0, _skinny_silence),
                   SkinnyFramePerPacketField('maxframes', 0),
                   LEIntField('unknown1', 0),
                   LEIntField('callid', 0)]

    def guess_payload_class(self, p):
        return conf.padding_layer


class SkinnyMessageCloseReceiveChannel(Packet):
    name = 'close receive channel'
    fields_desc = [LEIntField('conference', 0),
                   LEIntField('passthru', 0),
                   IPField('remote', '0.0.0.0'),
                   LEIntField('port', RandShort()),
                   SkinnyRateField('rate', 20),
                   LEIntEnumField('codec', 4, _skinny_codecs),
                   LEIntField('precedence', 200),
                   LEIntEnumField('silence', 0, _skinny_silence),
                   LEIntField('callid', 0)]


class SkinnyMessageStopMultiMediaTransmission(Packet):
    name = 'stop multimedia transmission'
    fields_desc = [LEIntField('conference', 0),
                   LEIntField('passthru', 0),
                   LEIntField('callid', 0)]


class Skinny(Packet):
    name = "Skinny"
    fields_desc = [LEIntField("len", None),
                   LEIntField("res", 0),
                   LEIntEnumField("msg", 0, skinny_messages_cls)]

    def post_build(self, pkt, p):
        if self.len is None:
            # on compte pas les headers len et reserved
            tmp_len = len(p) + len(pkt) - 8
            pkt = struct.pack('@I', tmp_len) + pkt[4:]
        return pkt + p

# An helper


def get_cls(name, fallback_cls):
    return globals().get(name, fallback_cls)
    # return __builtin__.__dict__.get(name, fallback_cls)


for msgid, strcls in skinny_messages_cls.items():
    cls = get_cls(strcls, SkinnyMessageGeneric)
    bind_layers(Skinny, cls, {"msg": msgid})

bind_layers(TCP, Skinny, {"dport": 2000})
bind_layers(TCP, Skinny, {"sport": 2000})

if __name__ == "__main__":
    from scapy.main import interact
    interact(mydict=globals(), mybanner="Welcome to Skinny add-on")



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy RPC
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

# ms-samr.idl compiled on 06/07/2025
# This file is a stripped version ! Use scapy-rpc for the full.
"""
RPC definitions for the following interfaces:
- samr (v1.0): 12345778-1234-ABCD-EF00-0123456789AC
This file is auto-generated by midl-to-scapy, do not modify.
"""

import uuid


from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfVarStrLenField,
    NDRConfVarStrLenFieldUtf16,
    NDRContextHandle,
    NDRFullEmbPointerField,
    NDRFullPointerField,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    register_dcerpc_interface,
)


class SamrConnect_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRShortField("ServerName", 0)),
        NDRIntField("DesiredAccess", 0),
    ]


class SamrConnect_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("ServerHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RPC_UNICODE_STRING(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRShortField("Length", None, size_of="Buffer", adjust=lambda _, x: (x * 2)),
        NDRShortField(
            "MaximumLength", None, size_of="Buffer", adjust=lambda _, x: (x * 2)
        ),
        NDRFullEmbPointerField(
            NDRConfVarStrLenFieldUtf16(
                "Buffer",
                "",
                size_is=lambda pkt: (pkt.MaximumLength // 2),
                length_is=lambda pkt: (pkt.Length // 2),
            )
        ),
    ]


class PSAMPR_RID_ENUMERATION(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("RelativeId", 0),
        NDRPacketField("Name", RPC_UNICODE_STRING(), RPC_UNICODE_STRING),
    ]


class PSAMPR_ENUMERATION_BUFFER(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("EntriesRead", None, size_of="Buffer"),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "Buffer",
                [PSAMPR_RID_ENUMERATION()],
                PSAMPR_RID_ENUMERATION,
                size_is=lambda pkt: pkt.EntriesRead,
            )
        ),
    ]


class SamrEnumerateDomainsInSamServer_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("ServerHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("EnumerationContext", 0),
        NDRIntField("PreferedMaximumLength", 0),
    ]


class SamrEnumerateDomainsInSamServer_Response(NDRPacket):
    fields_desc = [
        NDRIntField("EnumerationContext", 0),
        NDRFullPointerField(
            NDRPacketField(
                "Buffer", PSAMPR_ENUMERATION_BUFFER(), PSAMPR_ENUMERATION_BUFFER
            )
        ),
        NDRIntField("CountReturned", 0),
        NDRIntField("status", 0),
    ]


SAMR_OPNUMS = {
    0: DceRpcOp(SamrConnect_Request, SamrConnect_Response),
    # 4: Opnum4NotUsedOnWire,
    6: DceRpcOp(
        SamrEnumerateDomainsInSamServer_Request,
        SamrEnumerateDomainsInSamServer_Response,
    ),
    # 42: Opnum42NotUsedOnWire,
    # 43: Opnum43NotUsedOnWire,
    # 59: Opnum59NotUsedOnWire,
    # 60: Opnum60NotUsedOnWire,
    # 61: Opnum61NotUsedOnWire,
    # 63: Opnum63NotUsedOnWire,
    # 68: Opnum68NotUsedOnWire,
    # 69: Opnum69NotUsedOnWire,
    # 70: Opnum70NotUsedOnWire,
    # 71: Opnum71NotUsedOnWire,
    # 72: Opnum72NotUsedOnWire,
    # 75: Opnum75NotUsedOnWire,
    # 76: Opnum76NotUsedOnWire
}
register_dcerpc_interface(
    name="samr",
    uuid=uuid.UUID("12345778-1234-ABCD-EF00-0123456789AC"),
    version="1.0",
    opnums=SAMR_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Color themes for the interactive console.
"""

##################
#  Color themes  #
##################

import html
import sys

from typing import (
    Any,
    List,
    Optional,
    Tuple,
    cast,
)
from scapy.compat import Protocol


class ColorTable:
    colors = {  # Format: (ansi, pygments)
        # foreground
        "black": ("\033[30m", "#ansiblack"),
        "red": ("\033[31m", "#ansired"),
        "green": ("\033[32m", "#ansigreen"),
        "yellow": ("\033[33m", "#ansiyellow"),
        "blue": ("\033[34m", "#ansiblue"),
        "purple": ("\033[35m", "#ansipurple"),
        "cyan": ("\033[36m", "#ansicyan"),
        "white": ("\033[37m", "#ansiwhite"),
        "grey": ("\033[38;5;246m", "#ansiwhite"),
        "reset": ("\033[39m", "noinherit"),
        # background
        "bg_black": ("\033[40m", "bg:#ansiblack"),
        "bg_red": ("\033[41m", "bg:#ansired"),
        "bg_green": ("\033[42m", "bg:#ansigreen"),
        "bg_yellow": ("\033[43m", "bg:#ansiyellow"),
        "bg_blue": ("\033[44m", "bg:#ansiblue"),
        "bg_purple": ("\033[45m", "bg:#ansipurple"),
        "bg_cyan": ("\033[46m", "bg:#ansicyan"),
        "bg_white": ("\033[47m", "bg:#ansiwhite"),
        "bg_reset": ("\033[49m", "noinherit"),
        # specials
        "normal": ("\033[0m", "noinherit"),  # color & brightness
        "bold": ("\033[1m", "bold"),
        "uline": ("\033[4m", "underline"),
        "blink": ("\033[5m", ""),
        "invert": ("\033[7m", ""),
    }
    inv_map = {v[0]: v[1] for k, v in colors.items()}

    def __repr__(self):
        # type: () -> str
        return "<ColorTable>"

    def __getattr__(self, attr):
        # type: (str) -> str
        return self.colors.get(attr, [""])[0]

    def ansi_to_pygments(self, x):
        # type: (str) -> str
        """
        Transform ansi encoded text to Pygments text
        """
        for k, v in self.inv_map.items():
            x = x.replace(k, " " + v)
        return x.strip()


Color = ColorTable()


class _ColorFormatterType(Protocol):
    def __call__(self,
                 val: Any,
                 fmt: Optional[str] = None,
                 fmt2: str = "",
                 before: str = "",
                 after: str = "") -> str:
        pass


def create_styler(fmt=None,  # type: Optional[str]
                  before="",  # type: str
                  after="",  # type: str
                  fmt2="%s"  # type: str
                  ):
    # type: (...) -> _ColorFormatterType
    def do_style(val: Any,
                 fmt: Optional[str] = fmt,
                 fmt2: str = fmt2,
                 before: str = before,
                 after: str = after) -> str:
        if fmt is None:
            sval = str(val)
        else:
            sval = fmt % val
        return fmt2 % (before + sval + after)
    return do_style


class ColorTheme:
    style_normal = ""
    style_prompt = ""
    style_punct = ""
    style_id = ""
    style_not_printable = ""
    style_layer_name = ""
    style_field_name = ""
    style_field_value = ""
    style_emph_field_name = ""
    style_emph_field_value = ""
    style_depreciate_field_name = ""
    style_packetlist_name = ""
    style_packetlist_proto = ""
    style_packetlist_value = ""
    style_fail = ""
    style_success = ""
    style_odd = ""
    style_even = ""
    style_opening = ""
    style_active = ""
    style_closed = ""
    style_left = ""
    style_right = ""
    style_logo = ""

    def __repr__(self):
        # type: () -> str
        return "<%s>" % self.__class__.__name__

    def __reduce__(self):
        # type: () -> Tuple[type, Any, Any]
        return (self.__class__, (), ())

    def __getattr__(self, attr):
        # type: (str) -> _ColorFormatterType
        if attr in ["__getstate__", "__setstate__", "__getinitargs__",
                    "__reduce_ex__"]:
            raise AttributeError()
        return create_styler()

    def format(self, string, fmt):
        # type: (str, str) -> str
        for style in fmt.split("+"):
            string = getattr(self, style)(string)
        return string


class NoTheme(ColorTheme):
    pass


class AnsiColorTheme(ColorTheme):
    def __getattr__(self, attr):
        # type: (str) -> _ColorFormatterType
        if attr.startswith("__"):
            raise AttributeError(attr)
        s = "style_%s" % attr
        if s in self.__class__.__dict__:
            before = getattr(self, s)
            after = self.style_normal
        elif not isinstance(self, BlackAndWhite) and attr in Color.colors:
            before = Color.colors[attr][0]
            after = Color.colors["normal"][0]
        else:
            before = after = ""

        return create_styler(before=before, after=after)


class BlackAndWhite(AnsiColorTheme, NoTheme):
    pass


class DefaultTheme(AnsiColorTheme):
    style_normal = Color.normal
    style_prompt = Color.blue + Color.bold
    style_punct = Color.normal
    style_id = Color.blue + Color.bold
    style_not_printable = Color.white
    style_depreciate_field_name = Color.grey
    style_layer_name = Color.red + Color.bold
    style_field_name = Color.blue
    style_field_value = Color.purple
    style_emph_field_name = Color.blue + Color.uline + Color.bold
    style_emph_field_value = Color.purple + Color.uline + Color.bold
    style_packetlist_name = Color.red + Color.bold
    style_packetlist_proto = Color.blue
    style_packetlist_value = Color.purple
    style_fail = Color.red + Color.bold
    style_success = Color.blue + Color.bold
    style_even = Color.black + Color.bold
    style_odd = Color.black
    style_opening = Color.yellow
    style_active = Color.black
    style_closed = Color.white
    style_left = Color.blue + Color.invert
    style_right = Color.red + Color.invert
    style_logo = Color.green + Color.bold


class BrightTheme(AnsiColorTheme):
    style_normal = Color.normal
    style_punct = Color.normal
    style_id = Color.yellow + Color.bold
    style_layer_name = Color.red + Color.bold
    style_field_name = Color.yellow + Color.bold
    style_field_value = Color.purple + Color.bold
    style_emph_field_name = Color.yellow + Color.bold
    style_emph_field_value = Color.green + Color.bold
    style_packetlist_name = Color.red + Color.bold
    style_packetlist_proto = Color.yellow + Color.bold
    style_packetlist_value = Color.purple + Color.bold
    style_fail = Color.red + Color.bold
    style_success = Color.blue + Color.bold
    style_even = Color.black + Color.bold
    style_odd = Color.black
    style_left = Color.cyan + Color.invert
    style_right = Color.purple + Color.invert
    style_logo = Color.green + Color.bold


class RastaTheme(AnsiColorTheme):
    style_normal = Color.normal + Color.green + Color.bold
    style_prompt = Color.yellow + Color.bold
    style_punct = Color.red
    style_id = Color.green + Color.bold
    style_not_printable = Color.green
    style_layer_name = Color.red + Color.bold
    style_field_name = Color.yellow + Color.bold
    style_field_value = Color.green + Color.bold
    style_emph_field_name = Color.green
    style_emph_field_value = Color.green
    style_packetlist_name = Color.red + Color.bold
    style_packetlist_proto = Color.yellow + Color.bold
    style_packetlist_value = Color.green + Color.bold
    style_fail = Color.red
    style_success = Color.red + Color.bold
    style_even = Color.yellow
    style_odd = Color.green
    style_left = Color.yellow + Color.invert
    style_right = Color.red + Color.invert
    style_logo = Color.green + Color.bold


class ColorOnBlackTheme(AnsiColorTheme):
    """Color theme for black backgrounds"""
    style_normal = Color.normal
    style_prompt = Color.green + Color.bold
    style_punct = Color.normal
    style_id = Color.green
    style_not_printable = Color.black + Color.bold
    style_layer_name = Color.yellow + Color.bold
    style_field_name = Color.cyan
    style_field_value = Color.purple + Color.bold
    style_emph_field_name = Color.cyan + Color.bold
    style_emph_field_value = Color.red + Color.bold
    style_packetlist_name = Color.black + Color.bold
    style_packetlist_proto = Color.yellow + Color.bold
    style_packetlist_value = Color.purple + Color.bold
    style_fail = Color.red + Color.bold
    style_success = Color.green
    style_even = Color.black + Color.bold
    style_odd = Color.white
    style_opening = Color.yellow
    style_active = Color.white + Color.bold
    style_closed = Color.black + Color.bold
    style_left = Color.cyan + Color.bold
    style_right = Color.red + Color.bold
    style_logo = Color.green + Color.bold


class FormatTheme(ColorTheme):
    def __getattr__(self, attr: str) -> _ColorFormatterType:
        if attr.startswith("__"):
            raise AttributeError(attr)
        colfmt = self.__class__.__dict__.get("style_%s" % attr, "%s")
        return create_styler(fmt2=colfmt)


class LatexTheme(FormatTheme):
    r"""
    You can prepend the output from this theme with
    \tt\obeyspaces\obeylines\tiny\noindent
    """
    style_prompt = r"\textcolor{blue}{%s}"
    style_not_printable = r"\textcolor{gray}{%s}"
    style_layer_name = r"\textcolor{red}{\bf %s}"
    style_field_name = r"\textcolor{blue}{%s}"
    style_field_value = r"\textcolor{purple}{%s}"
    style_emph_field_name = r"\textcolor{blue}{\underline{%s}}"  # ul
    style_emph_field_value = r"\textcolor{purple}{\underline{%s}}"  # ul
    style_packetlist_name = r"\textcolor{red}{\bf %s}"
    style_packetlist_proto = r"\textcolor{blue}{%s}"
    style_packetlist_value = r"\textcolor{purple}{%s}"
    style_fail = r"\textcolor{red}{\bf %s}"
    style_success = r"\textcolor{blue}{\bf %s}"
    style_left = r"\textcolor{blue}{%s}"
    style_right = r"\textcolor{red}{%s}"
#    style_even = r"}{\bf "
#    style_odd = ""
    style_logo = r"\textcolor{green}{\bf %s}"

    def __getattr__(self, attr: str) -> _ColorFormatterType:
        from scapy.utils import tex_escape
        styler = super(LatexTheme, self).__getattr__(attr)
        return cast(
            _ColorFormatterType,
            lambda x, *args, **kwargs: styler(tex_escape(str(x)), *args, **kwargs),
        )


class LatexTheme2(FormatTheme):
    style_prompt = r"@`@textcolor@[@blue@]@@[@%s@]@"
    style_not_printable = r"@`@textcolor@[@gray@]@@[@%s@]@"
    style_layer_name = r"@`@textcolor@[@red@]@@[@@`@bfseries@[@@]@%s@]@"
    style_field_name = r"@`@textcolor@[@blue@]@@[@%s@]@"
    style_field_value = r"@`@textcolor@[@purple@]@@[@%s@]@"
    style_emph_field_name = r"@`@textcolor@[@blue@]@@[@@`@underline@[@%s@]@@]@"
    style_emph_field_value = r"@`@textcolor@[@purple@]@@[@@`@underline@[@%s@]@@]@"  # noqa: E501
    style_packetlist_name = r"@`@textcolor@[@red@]@@[@@`@bfseries@[@@]@%s@]@"
    style_packetlist_proto = r"@`@textcolor@[@blue@]@@[@%s@]@"
    style_packetlist_value = r"@`@textcolor@[@purple@]@@[@%s@]@"
    style_fail = r"@`@textcolor@[@red@]@@[@@`@bfseries@[@@]@%s@]@"
    style_success = r"@`@textcolor@[@blue@]@@[@@`@bfseries@[@@]@%s@]@"
    style_even = r"@`@textcolor@[@gray@]@@[@@`@bfseries@[@@]@%s@]@"
#    style_odd = r"@`@textcolor@[@black@]@@[@@`@bfseries@[@@]@%s@]@"
    style_left = r"@`@textcolor@[@blue@]@@[@%s@]@"
    style_right = r"@`@textcolor@[@red@]@@[@%s@]@"
    style_logo = r"@`@textcolor@[@green@]@@[@@`@bfseries@[@@]@%s@]@"


class HTMLTheme(FormatTheme):
    style_prompt = "<span class=prompt>%s</span>"
    style_not_printable = "<span class=not_printable>%s</span>"
    style_layer_name = "<span class=layer_name>%s</span>"
    style_field_name = "<span class=field_name>%s</span>"
    style_field_value = "<span class=field_value>%s</span>"
    style_emph_field_name = "<span class=emph_field_name>%s</span>"
    style_emph_field_value = "<span class=emph_field_value>%s</span>"
    style_packetlist_name = "<span class=packetlist_name>%s</span>"
    style_packetlist_proto = "<span class=packetlist_proto>%s</span>"
    style_packetlist_value = "<span class=packetlist_value>%s</span>"
    style_fail = "<span class=fail>%s</span>"
    style_success = "<span class=success>%s</span>"
    style_even = "<span class=even>%s</span>"
    style_odd = "<span class=odd>%s</span>"
    style_left = "<span class=left>%s</span>"
    style_right = "<span class=right>%s</span>"


class HTMLTheme2(HTMLTheme):
    style_prompt = "#[#span class=prompt#]#%s#[#/span#]#"
    style_not_printable = "#[#span class=not_printable#]#%s#[#/span#]#"
    style_layer_name = "#[#span class=layer_name#]#%s#[#/span#]#"
    style_field_name = "#[#span class=field_name#]#%s#[#/span#]#"
    style_field_value = "#[#span class=field_value#]#%s#[#/span#]#"
    style_emph_field_name = "#[#span class=emph_field_name#]#%s#[#/span#]#"
    style_emph_field_value = "#[#span class=emph_field_value#]#%s#[#/span#]#"
    style_packetlist_name = "#[#span class=packetlist_name#]#%s#[#/span#]#"
    style_packetlist_proto = "#[#span class=packetlist_proto#]#%s#[#/span#]#"
    style_packetlist_value = "#[#span class=packetlist_value#]#%s#[#/span#]#"
    style_fail = "#[#span class=fail#]#%s#[#/span#]#"
    style_success = "#[#span class=success#]#%s#[#/span#]#"
    style_even = "#[#span class=even#]#%s#[#/span#]#"
    style_odd = "#[#span class=odd#]#%s#[#/span#]#"
    style_left = "#[#span class=left#]#%s#[#/span#]#"
    style_right = "#[#span class=right#]#%s#[#/span#]#"


def apply_ipython_style(shell):
    # type: (Any) -> None
    """Updates the specified IPython console shell with
    the conf.color_theme scapy theme."""
    try:
        from IPython.terminal.prompts import Prompts, Token
    except Exception:
        from scapy.error import log_loading
        log_loading.warning(
            "IPython too old. Shell color won't be handled."
        )
        return
    from scapy.config import conf
    scapy_style = {}
    # Overwrite colors
    if isinstance(conf.color_theme, NoTheme):
        shell.colors = 'nocolor'
    elif isinstance(conf.color_theme, BrightTheme):
        # lightbg is optimized for light backgrounds
        shell.colors = 'lightbg'
    elif isinstance(conf.color_theme, ColorOnBlackTheme):
        # linux is optimised for dark backgrounds
        shell.colors = 'linux'
    else:
        # default
        shell.colors = 'neutral'
    try:
        get_ipython()  # type: ignore
        # This function actually contains tons of hacks
        color_magic = shell.magics_manager.magics["line"]["colors"]
        color_magic(shell.colors)
    except NameError:
        pass
    # Prompt Style
    if isinstance(conf.prompt, Prompts):
        # Set custom prompt style
        shell.prompts_class = conf.prompt
    else:
        if isinstance(conf.color_theme, (FormatTheme, NoTheme)):
            # Formatable
            if isinstance(conf.color_theme, HTMLTheme):
                prompt = html.escape(conf.prompt)
            elif isinstance(conf.color_theme, LatexTheme):
                from scapy.utils import tex_escape
                prompt = tex_escape(conf.prompt)
            else:
                prompt = conf.prompt
            prompt = conf.color_theme.prompt(prompt)
        else:
            # Needs to be manually set
            prompt = str(conf.prompt)
            scapy_style[Token.Prompt] = Color.ansi_to_pygments(
                conf.color_theme.style_prompt
            )

        class ClassicPrompt(Prompts):
            def in_prompt_tokens(self, cli=None):
                # type: (Any) -> List[Tuple[Any, str]]
                return [(Token.Prompt, prompt), ]

            def out_prompt_tokens(self):
                # type: () -> List[Tuple[Any, str]]
                return [(Token.OutPrompt, ''), ]
        # Apply classic prompt style
        shell.prompts_class = ClassicPrompt
        sys.ps1 = prompt
    # Register scapy color style
    shell.highlighting_style_overrides = scapy_style
    # Apply if Live
    try:
        get_ipython().refresh_style()  # type: ignore
    except NameError:
        pass
